from typing import List
import os
import sys
import subprocess
import json

from vosk import Model, KaldiRecognizer
from pydub import AudioSegment
from tqdm import tqdm

from .models import load_model
from ..audio import get_audiofile_length




def transcribe_segment(segment: AudioSegment) -> str:
    """ Transcribe a short AudioSegment """
    assert segment.frame_rate == 16000, f"Wrong sample rate {segment.frame_rate=}"
    assert segment.sample_width == 2, f"Wrong sample width {segment.sample_width=}"
    assert segment.channels == 1, f"Wrong number of channels {segment.channels=}"

    model = load_model()
    recognizer = KaldiRecognizer(model, 16000)
    recognizer.SetWords(True)
    
    data = segment.raw_data
    text = []
    i = 0
    while i + 4000 < len(data):
        if recognizer.AcceptWaveform(data[i:i+4000]):
            text.append(json.loads(recognizer.Result())["text"])
        i += 4000
    recognizer.AcceptWaveform(data[i:])
    text.append(json.loads(recognizer.FinalResult())["text"])

    return text



def transcribe_segment_timecoded(segment: AudioSegment) -> List[dict]:
    """ Transcribe a short AudioSegment, keeping the timecodes

        The resulting transcription is a list of Vosk tokens
        Each Vosk token is a dictionary of the form:
            {'word': str, 'start': float, 'end': float, 'conf': float}
        'start' and 'end' keys are in seconds
        'conf' is a normalized confidence score
    """
    assert segment.frame_rate == 16000
    assert segment.sample_width == 2
    assert segment.channels == 1
    
    model = load_model()
    recognizer = KaldiRecognizer(model, 16000)
    recognizer.SetWords(True)
    
    data = segment.get_array_of_samples().tobytes()
    timecoded_text = []
    i = 0
    while i + 4000 < len(data):
        if recognizer.AcceptWaveform(data[i:i+4000]):
            result = json.loads(recognizer.Result())
            if "result" in result:
                timecoded_text.extend(result["result"])
        i += 4000
    recognizer.AcceptWaveform(data[i:])
    result = json.loads(recognizer.FinalResult())
    if "result" in result:
        timecoded_text.extend(result["result"])
    return timecoded_text



def transcribe_segment_timecoded_callback(segment: AudioSegment, callback: callable):
    """ Transcribe a short AudioSegment, keeping the timecodes,
        Send result to callback function for every detected utterances

        The resulting transcription is a list of Vosk tokens
        Each Vosk token is a dictionary of the form:
            {'word': str, 'start': float, 'end': float, 'conf': float}
        'start' and 'end' keys are in seconds
        'conf' is a normalized confidence score
    """
    assert segment.frame_rate == 16000, f"Wrong sampling rate {segment.frame_rate=} (should be 16000)"
    assert segment.sample_width == 2, f"Wrong sample width {segment.sample_width=} (should be 2)"
    assert segment.channels == 1, f"Wrong number of channels {segment.channels=} (should be 1)"
    
    model = load_model()
    recognizer = KaldiRecognizer(model, 16000)
    recognizer.SetWords(True)
    
    data = segment.get_array_of_samples().tobytes()
    i = 0
    while i + 4000 < len(data):
        if recognizer.AcceptWaveform(data[i:i+4000]):
            result = json.loads(recognizer.Result())
            if "result" in result:
                callback(result["result"])
        i += 4000
    recognizer.AcceptWaveform(data[i:])
    result = json.loads(recognizer.FinalResult())
    if "result" in result:
        callback(result["result"])



def transcribe_file(filepath: str) -> List[str]:
    if not os.path.exists(filepath):
        print("Couldn't find {}".format(filepath), file=sys.stderr)

    model = load_model()
    recognizer = KaldiRecognizer(model, 16000)
    recognizer.SetWords(True)
    
    text = []

    with subprocess.Popen(["ffmpeg", "-loglevel", "quiet", "-i",
                                filepath,
                                "-ar", "16000" , "-ac", "1", "-f", "s16le", "-"],
                                stdout=subprocess.PIPE) as process:

        while True:
            data = process.stdout.read(4000)
            if len(data) == 0:
                break
            if recognizer.AcceptWaveform(data):
                sentence = json.loads(recognizer.Result())["text"]
                if sentence:
                    text.append(sentence)
        sentence = json.loads(recognizer.FinalResult())["text"]
        if sentence:
            text.append(sentence)
    
    return text



def transcribe_file_timecoded(filepath: str, show_progress_bar=True) -> List[dict]:
    """ Return list of infered words with associated timecodes (vosk format)

        The resulting transcription is a list of Vosk tokens
        Each Vosk token is a dictionary of the form:
            {'word': str, 'start': float, 'end': float, 'conf': float}
        'start' and 'end' keys are in seconds
        'conf' is a normalized confidence score
    """

    def format_output(result) -> List[dict]:
        jres = json.loads(result)
        return jres.get("result", [])

    model = load_model()
    recognizer = KaldiRecognizer(model, 16000)
    recognizer.SetWords(True)

    total_duration = get_audiofile_length(filepath)
    progress = 0.0

    if show_progress_bar:
        progress_bar = tqdm(total=total_duration, unit='s', unit_scale=True)
    
    tokens = []
    with subprocess.Popen(["ffmpeg", "-loglevel", "quiet", "-i",
                            filepath,
                            "-ar", "16000" , "-ac", "1", "-f", "s16le", "-"],
                            stdout=subprocess.PIPE) as process:

        while True:
            data = process.stdout.read(4000)
            if len(data) == 0:
                break
            if recognizer.AcceptWaveform(data):
                tokens.extend(format_output(recognizer.Result()))

            progress += (len(data) // 2) / 16000

            if show_progress_bar:
                progress_bar.n = min(progress, total_duration)
                progress_bar.refresh()
        
        tokens.extend(format_output(recognizer.FinalResult()))
    if show_progress_bar:
        progress_bar.close()
    
    return tokens


"""
Load various dictionaries from local resources

The option should be given to load user dictionaries as well, stored in a system folder
"""

import sys
# import os
# import platform
import importlib.resources


# dict_root = importlib.resources.files("anaouder") / "dicts"

# print(f"loading dicts in {dict_root}", file=sys.stderr)


# Proper nouns dictionary
# with phonemes when name has a foreign or particular pronunciations

def load_proper_nouns():
    proper_nouns = dict()
    proper_nouns_files = [
        "proper_nouns_phon.tsv",
        "places.tsv",
        # "countries.tsv",
        "last_names.tsv",
        "first_names.tsv",
    ]

    # print(f"Loading proper nouns dictionaries", file=sys.stderr)
    
    for file in proper_nouns_files:
        try:
            # Use importlib.resources to access files within the package
            with importlib.resources.files("anaouder.dicts").joinpath(file).open('r', encoding='utf-8') as f:
                for l in f.readlines():
                    l = l.strip()
                    if l.startswith('#') or not l: continue
                    w, *pron = l.split(maxsplit=1)
                    pron = pron or []
                    
                    if w in proper_nouns and pron:
                        if pron[0] not in proper_nouns[w]: # Avoid duplicate entries, which Kaldi hates
                            proper_nouns[w].append(pron[0])
                    else:
                        proper_nouns[w] = pron
        except FileNotFoundError:
            print(f"Missing dictionary file {file}", file=sys.stderr)
            continue
    
    return proper_nouns

proper_nouns = load_proper_nouns()



# Nouns dictionary
# Things that you can count

def load_nouns_f():
    nouns_f = set()
    filepath = "noun_f.tsv"

    try:
        with importlib.resources.files("anaouder.dicts").joinpath(filepath).open('r', encoding='utf-8') as f:
            
            for l in f.readlines():
                l = l.strip()
                if l.startswith('#') or not l: continue
                nouns_f.add(l)
        
        return nouns_f
    except FileNotFoundError:
        print(f"Missing dictionary file {filepath}", file=sys.stderr)

nouns_f = load_nouns_f()



def load_nouns_m():
    nouns_m = set()
    filepath = "noun_m.tsv"

    try:
        with importlib.resources.files("anaouder.dicts").joinpath("noun_m.tsv").open('r', encoding='utf-8') as f:
            for l in f.readlines():
                l = l.strip()
                if l.startswith('#') or not l: continue
                nouns_m.add(l)
        
        return nouns_m
    except FileNotFoundError:
        print(f"Missing dictionary file {filepath}", file=sys.stderr)

nouns_m = load_nouns_m()



# Acronyms dictionary

def load_acronyms():
    """
    Acronyms are stored in UPPERCASE in dictionary
    Values are lists of strings for all possible pronunciation of an acronym
    """
    acronyms = dict()
    filepath = "acronyms.tsv"

    # for l in "BCDFGHIJKLMPQRSTUVWXZ":
    #     acronyms[l] = [acr2f[l]]
    
    try:
        with importlib.resources.files("anaouder.dicts").joinpath(filepath).open('r', encoding='utf-8') as f:
            for l in f.readlines():
                l = l.strip()
                if l.startswith('#') or not l: continue
                acr, pron = l.split(maxsplit=1)
                if acr in acronyms:
                    acronyms[acr].append(pron)
                else:
                    acronyms[acr] = [pron]
    except FileNotFoundError:
        print(f"Missing dictionary file {filepath}", file=sys.stderr)
    
    return acronyms

acronyms = load_acronyms()



# Abbreviations

def load_abbreviations():
    abbreviations = dict()
    filepath = "abbreviations.tsv"

    try:
        with importlib.resources.files("anaouder.dicts").joinpath(filepath).open('r', encoding='utf-8') as f:
            for l in f.readlines():
                l = l.strip()
                if l.startswith('#') or not l: continue
                k, v = l.split('\t')
                # v = v.split()
                abbreviations[k] = v
    except FileNotFoundError:
        print(f"Missing dictionary file {filepath}", file=sys.stderr)
    
    return abbreviations

abbreviations = load_abbreviations()


# Interjections

def load_interjections():
    """
    Acronyms are stored in UPPERCASE in dictionary
    Values are lists of strings for all possible pronunciation of an acronym
    """
    interjections = dict()
    filepath = "interjections.tsv"
    
    try:
        with importlib.resources.files("anaouder.dicts").joinpath(filepath).open('r', encoding='utf-8') as f:
            for l in f.readlines():
                l = l.strip()
                if l.startswith('#') or not l: continue
                interj, *pron = l.split(maxsplit=1)
                pron = pron if pron else []

                if interj in interjections and pron:
                    interjections[interj].append(pron)
                else:
                    interjections[interj] = pron
    except FileNotFoundError:
        print(f"Missing dictionary file {filepath}", file=sys.stderr)

    return interjections

interjections = load_interjections()


# Common word mistakes

def load_corrected_tokens():
    corrected_tokens = dict()
    filepath = "corrected_tokens.tsv"

    try:
        with importlib.resources.files("anaouder.dicts").joinpath(filepath).open('r', encoding='utf-8') as f:
            for l in f.readlines():
                l = l.strip()
                if l.startswith('#') or not l: continue
                k, v = l.split('\t')
                v = v.split()
                corrected_tokens[k] = v
    except FileNotFoundError:
        print(f"Missing dictionary file {filepath}", file=sys.stderr)
    
    return corrected_tokens

corrected_tokens = load_corrected_tokens()


# Standardization tokens

def load_standard_tokens():
    standard_tokens = dict()
    filepath = "standard_tokens.tsv"

    try:
        with importlib.resources.files("anaouder.dicts").joinpath(filepath).open('r', encoding='utf-8') as f:
            for l in f.readlines():
                l = l.strip()
                if l.startswith('#') or not l: continue
                k, v = l.split('\t')
                v = v.split()
                standard_tokens[k] = v
    except FileNotFoundError:
        print(f"Missing dictionary file {filepath}", file=sys.stderr)
    
    return standard_tokens

standard_tokens = load_standard_tokens()



from typing import List, Optional

import sys
import subprocess
import json
import os.path
from os import listdir
from random import choice
from tempfile import NamedTemporaryFile

from colorama import Fore

from pydub import AudioSegment
from pydub.utils import get_player_name
from pydub.generators import WhiteNoise


AUDIO_FORMATS = ('wav', 'mp3', 'm4a', 'ogg', 'mp4', 'mkv')


def find_associated_audiofile(path: str, silent=False) -> Optional[str]:
    basename, _ = os.path.splitext(path)

    for ext in AUDIO_FORMATS:
        audio_path = os.path.extsep.join((basename, ext))
        if os.path.exists(audio_path):
            if not silent:
                print("Found audio file:", audio_path)
            return audio_path
    print(Fore.RED + f"Could not find {audio_path}" + Fore.RESET, file=sys.stderr)
    return None



def load_audiofile(path: str, sr=16000) -> AudioSegment:
    data = AudioSegment.from_file(path)
    data = prepare_segment_for_decoding(data)
    return data



def get_audio_segment(i, audio: AudioSegment, segments):
    """
    Args:
        i (int) : an index
        audio (AudioSegment)
        segments: list of tuple of floats (seconds)
    """
    start = int(segments[i][0] * 1000)
    stop = int(segments[i][1] * 1000)
    seg = audio[start: stop]
    return seg



def audio_segments(audio, segments):
    for start, stop in segments:
        yield audio[start, stop]



def prepare_segment_for_decoding(segment: AudioSegment) -> AudioSegment:
    """ Ensure that the segment is the right sampling rate and depth """

    if segment.channels > 1:
        segment = segment.set_channels(1)
    if segment.frame_rate != 16000:
        segment = segment.set_frame_rate(16000)
    if segment.sample_width != 2:
        segment = segment.set_sample_width(2)
    return segment



def play_audio_segment(i, song, segments, speed):
    play_with_ffplay(get_audio_segment(i, song, segments), speed)



def get_audiofile_info(filename) -> dict:
    r = subprocess.check_output(['ffprobe', '-hide_banner', '-v', 'panic', '-show_streams', '-of', 'json', filename])
    r = json.loads(r)
    return r['streams'][0]



def get_audiofile_length(filename) -> float:
    """ Get audio file length in seconds """
    info = get_audiofile_info(filename)
    if "duration" in info:
        return float(info["duration"])
    elif "tags" in info:
        # For MKV files
        h, m, s = info["tags"]["DURATION"].split(':')
        return float(h) * 3600 + float(m) * 60 + float(s)



def is_audiofile_valid_format(filename) -> bool:
    """ Returns True if audio file is 16KHz s16le PCM """
    info = get_audiofile_info(filename)
    if info["codec_name"] != "pcm_s16le":
        return False
    if info["sample_rate"] != "16000":
        return False
    if info["channels"] != 1:
        return False
    return True


    
def play_with_ffplay(seg, speed=1.0):
    with NamedTemporaryFile("w+b", suffix=".wav") as f:
        seg.export(f.name, "wav")
        player = get_player_name()
        p = subprocess.Popen(
            [player, "-nodisp", "-autoexit", "-loglevel", "quiet", "-af", f"atempo={speed}", f.name],
        )
        p.wait()



def convert_to_wav(src, dst, verbose=True, keep_orig=True):
    """
        Convert to 16kHz s16le mono PCM

        Parameters
        ----------
            verbose (bool):
                Info text to stdout
            keep_orig (bool):
                keep  original file (or rename if src and dst are the same)
    """
    src = os.path.abspath(src)
    dst = os.path.abspath(dst)
    if src == dst:
        # Rename original audio file (to keep it)
        rep, filename = os.path.split(src)
        basename, ext = os.path.splitext(filename)
        new_name = basename + "_orig" + ext
        new_src = os.path.join(rep, new_name)
        if verbose: print(Fore.YELLOW + f"AUDIO_CONV: renaming {filename} to {new_name}" + Fore.RESET)
        os.rename(src, new_src)
        src = new_src

    if verbose:
        print(Fore.YELLOW + f"AUDIO_CONV: converting '{src}' to '{dst}'..." + Fore.RESET)
    rep, filename = os.path.split(dst)
    dst = os.path.join(rep, filename)
    subprocess.call([
            'ffmpeg', '-v',
            'panic',
            '-i', src,
            '-acodec', 'pcm_s16le',
            '-ac', '1',
            '-ar', '16000',
            dst
        ])

    if not keep_orig:
        if verbose:
            print(Fore.YELLOW + f"AUDIO_CONV: Removing {src}" + Fore.RESET)
        os.remove(src)



def convert_to_mp3(src, dst, verbose=True, keep_orig=True):
    """ Convert to MP3 """
    if verbose:
        print(Fore.YELLOW + f"AUDIO_CONV: converting '{src}' to '{dst}'..." + Fore.RESET)
        
    if os.path.abspath(src) == os.path.abspath(dst):
        print(Fore.RED + "ERROR: source and destination are the same, skipping" + Fore.RESET)
        return -1
    rep, filename = os.path.split(dst)
    dst = os.path.join(rep, filename)
    subprocess.call(['ffmpeg', '-v', 'panic',
                     '-i', src,
                     '-ac', '1', dst])
    
    if not keep_orig:
        if verbose:
            print(Fore.YELLOW + f"AUDIO_CONV: Removing {src}")
        os.remove(src)



def concatenate_audiofiles(file_list, out_filename, remove=False):
    """ Concatenate a list of audio files to a single audio file

        Parameters
        ----------
            remove (bool):
                remove original files
    """

    if len(file_list) <= 1:
        return
    
    file_list_filename = "audiofiles.txt"
    with open(file_list_filename, 'w', encoding='utf-8') as f:
        f.write('\n'.join([f"file '{wav}'" for wav in file_list]))
    
    subprocess.call(['ffmpeg', '-v', 'panic',
                     '-f', 'concat',
                     '-safe', '0',
                     '-i', file_list_filename,
                     '-c', 'copy', out_filename])
    os.remove(file_list_filename)
    
    if remove:
        for fname in file_list:
            os.remove(fname)



def export_segment(audio_path: str, start: float, end: float, new_path: str):
    """ Split an audiofile in many segments

        Arguments:
            audio_path (str): the path of the original audiofile
            start (float) : time offset of beginning of segment
            end (float) : time offset of end of segment
            new_path (str): path of output file
    """
    # Cut the audio into segments using FFmpeg and suppress output
    subprocess.run([
        "ffmpeg",
        "-i", audio_path,
        "-ss", str(start),
        "-to", str(end),
        "-c", "copy", # Write using the same codec
        new_path
    ], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)



def get_min_max_energy(segment: AudioSegment, chunk_size=100, overlap=50):
    """Return the min and max RMS energy value for this audio segment"""
    min_energy = segment.max_possible_amplitude
    max_energy = 0
    for i in range(0, len(segment), chunk_size-overlap):
        chunk = segment[i: i+chunk_size]  # Pydub will take care of overflow
        mean = chunk.rms
        if mean < min_energy:
            min_energy = mean
        if mean > max_energy:
            max_energy = mean
    return (min_energy, max_energy)



def binary_split(audio: AudioSegment, treshold_ratio=0.1):
    min_e, max_e = get_min_max_energy(audio)
    delta_e = max_e - min_e
    thresh = min_e + delta_e * treshold_ratio

    # Find segments above the energy threshold
    chunk_size = 100 # millisec
    overlap = 50 # millisec
    segments = []
    seg_start = 0 # millisec
    in_noisy_chunk = False
    for i in range(0, len(audio), chunk_size-overlap):
        chunk = audio[i: i+chunk_size]  # Pydub will take care of overflow
        if chunk.rms >= thresh:
            if not in_noisy_chunk:
                seg_start = i
                in_noisy_chunk = True
        else:
            if in_noisy_chunk:
                segments.append((seg_start, i+chunk_size))
                in_noisy_chunk = False
    if in_noisy_chunk:
        segments.append((seg_start, len(audio)))

    if not segments:
        return segments
    # Find longest silence between segments
    max_sil_length = 0
    max_sil_length_idx = -1
    for i in range(1, len(segments)):
        sil_length = segments[i][0] - segments[i-1][1]
        if sil_length > max_sil_length:
            max_sil_length = sil_length
            max_sil_length_idx = i

    if max_sil_length_idx > 0:
        left_seg = (segments[0][0], segments[max_sil_length_idx-1][1])
        right_seg = (segments[max_sil_length_idx][0], segments[-1][1])
        return [left_seg, right_seg]
    return []



def split_to_segments(audio: AudioSegment, max_length=10, threshold_ratio=0.1) -> List:
    """
    Return a list of shorter sub-segments from a pydub AudioSegment.

    Sub-segments are represented as 2-elements lists [start, end]
    where 'start' and 'end' are in milliseconds.

    Parameters
    ----------
        max_length (float):
            sub-segments maximum length (in seconds)
        threshold_ratio (0.0 < float < 1.0):
            the silence threshold (depending on min/max energy of the audio segment)
    """
    segments_stack = [(0, len(audio))]
    short_segments = []

    while segments_stack:
        segment = segments_stack.pop()
        start, end = segment
        seg_len = end - start
        if seg_len <= max_length * 1000:
            short_segments.append(segment)
            continue
        sub_segments = binary_split(audio[start:end], threshold_ratio)
        segments_stack.extend([ (start + s, start + e) for s, e in sub_segments ])
    
    return sorted(short_segments)


#! /usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Programming pattern inspired by https://github.com/mideind/Tokenizer
"""


from typing import Iterator, Iterable, List, Any, Union, Set
from enum import Enum
import os.path
import re

from sentence_splitter import SentenceSplitter, split_text_into_sentences

from .definitions import (
    re_word, is_word, is_word_inclusive, re_extended_word,
    is_roman_number, is_ordinal, is_roman_ordinal,
    is_noun, is_noun_f, is_noun_m, is_proper_noun,
    is_time, match_time,
    is_unit_number, match_unit_number,
    PUNCTUATION, LETTERS, SI_UNITS,
    OPENING_QUOTES, CLOSING_QUOTES,
    PUNCT_PAIRS, OPENING_PUNCT, CLOSING_PUNCT,
)
from .utils import capitalize, is_capitalized
from ..dicts import acronyms, abbreviations, corrected_tokens, standard_tokens




class Token:
    # Enum
    RAW = -1
    SPECIAL_TOKEN = 0
    END_OF_SENTENCE = 1
    PUNCTUATION = 2
    WORD = 3
    NUMBER = 4
    ROMAN_NUMBER = 5
    NOUN = 6
    PROPER_NOUN = 7
    VERB = 8
    ACRONYM = 9
    ORDINAL = 10
    ROMAN_ORDINAL = 11
    TIME = 12
    UNIT = 13         # %, m, km2, kg...
    QUANTITY = 14     # a number and a unit (%, m, km2, kg, bloaz, den...)
    ABBREVIATIION = 15
    UNKNOWN = 99

    descr = {
        RAW: "RAW",
        SPECIAL_TOKEN: "SPECIAL_TOKEN",
        END_OF_SENTENCE: "END_OF_SENTENCE",
        WORD: "WORD",
        NUMBER: "NUMBER",
        ROMAN_NUMBER: "ROMAN_NUMBER",
        PUNCTUATION: "PUNCTUATION",
        NOUN: "NOUN",
        PROPER_NOUN: "PROPER_NOUN",
        VERB: "VERB",
        ACRONYM: "ACRONYM",
        ORDINAL: "ORDINAL",
        ROMAN_ORDINAL: "ROMAN_ORDINAL",
        TIME: "TIME",
        UNIT: "UNIT",
        QUANTITY: "QUANTITY",
        # PERCENT: "PERCENT",
        ABBREVIATIION: "ABBREVIATION",
        UNKNOWN: "UNKNOWN",
    }

    def __init__(self, data: str, kind: int=RAW, *flags):
        self.data = data
        self.norm = []
        self.kind = kind
        self.flags: Set[Flag] = set(flags)
        self.next = None # The token following this one, after "generate_tokens_lookahead" is called
    
    def __repr__(self):
        flags_name = [flag.name for flag in self.flags]
        return "Token(" + \
            f"{repr(self.data)}, {self.descr[self.kind]}" + \
            f", {flags_name if flags_name else ''}" + \
            ")"


class Flag(Enum):
    FIRST_WORD = 1
    MASCULINE = 2
    FEMININE = 3
    INCLUSIVE = 4
    CAPITALIZED = 5
    CORRECTED = 6
    OPENING_PUNCT = 7
    CLOSING_PUNCT = 8



_root = os.path.dirname(os.path.abspath(__file__))
_moses_prefix_file = os.path.join(_root, "moses_br.txt")

def split_sentences(text_or_gen: Union[str, Iterable[str]]) -> Iterator[str]:
    """ Split a line (or list of lines) according to its punctuation
        This function can be used independently
    """
   #print(text_or_gen)
    if isinstance(text_or_gen, str):
        text = text_or_gen
    else:
        text = ' '.join([line.strip() for line in text_or_gen])
    
    return split_text_into_sentences(
            text=text,
            language='br',
            non_breaking_prefix_file=_moses_prefix_file
        )


def split_sentences_old(text_or_gen: Union[str, Iterable[str]], **options: Any) -> Iterator[str]:
    """ Split a line (or list of lines) according to its punctuation
        This function can be used independently

        Parameters
        ----------
            end : str
                End the sentences with the given character

    """

    end_char = options.pop("end", '\n')
    # preserve_newline = options.pop("preserve_newline", False)

    if isinstance(text_or_gen, str):
        if not text_or_gen:
            return
        text_or_gen = [text_or_gen]
    
    token_stream = generate_raw_tokens(text_or_gen)
    token_stream = parse_punctuation(token_stream)
    token_stream = generate_tokens_lookahead(token_stream) # Needed to generate subsequent eos tokens
    token_stream = generate_eos_tokens(token_stream)

    current_sentence = []
    for tok in token_stream:
        if tok.kind == Token.END_OF_SENTENCE:
            yield detokenize(current_sentence, **options) + end_char
            current_sentence = []
        else:
            current_sentence.append(tok)
    if current_sentence:
        yield detokenize(current_sentence, **options) + end_char





def tokenize(text_or_gen: Union[str, Iterable[str]], **options: Any) -> Iterator[Token]:
    """
        Parameters
        ----------
            autocorrect: boolean
                Try to correct typos with a substitution dictionary
            
            norm_punct: boolean
                Normalize punctuation
        
        TODO:
            * &
    """

    # Arg options
    autocorrect = options.pop('autocorrect', False)
    #standardize = options.pop('standardize', False)
    
    token_stream = generate_raw_tokens(text_or_gen)
    token_stream = parse_punctuation(token_stream, **options)
    token_stream = generate_tokens_lookahead(token_stream)
    token_stream = generate_eos_tokens(token_stream)
    if autocorrect:
        token_stream = correct_tokens(token_stream)
    token_stream = parse_numerals(token_stream)
    token_stream = parse_regular_words(token_stream, **options)
    # token_stream = parse_acronyms(token_stream)

    return token_stream



def detokenize(token_stream: Iterator[Token], **options: Any) -> str:
    """
        Parameters
        ----------
            capitalize: boolean
                Capitalize sentence
            
            end: str
                String to append at end of each sentence
    """
    
    # Parse options
    end_sentence = options.pop('end', '')
    capitalize_opt = options.pop('capitalize', False)
    normalize = options.pop("normalize", False)
    # colored = options.pop("colored", False)

    parts: List[str] = []
    punct_stack = [] # Used to keep track of coupled punctuation (quotes and brackets)
    capitalize_next_word = capitalize_opt

    for tok in token_stream:
        data = tok.norm[0] if (normalize and tok.norm) else tok.data

        if capitalize_next_word:
            data = data.capitalize()
            capitalize_next_word = False

        prefix = ''
        if tok.kind == Token.PUNCTUATION:
            if data in '!?:;–':
                prefix = '\xa0' # Non-breakable space
            elif data == '"':
                if punct_stack and punct_stack[-1] == '"':
                    punct_stack.pop()
                    prefix = ''
                else:
                    punct_stack.append('"')
                    prefix = ' '
            elif data in OPENING_QUOTES:
                # we use a single '"' char to represent every kind of quotation mark
                # this prevents problems when mixing types of quotation marks
                punct_stack.append('"')
                prefix = ' '
            elif data in CLOSING_QUOTES:
                if punct_stack and punct_stack[-1] == '"':
                    punct_stack.pop()
                prefix = '\xa0' if data == '»' else ''
            elif data in '([':
                punct_stack.append(data)
                prefix = ' '
            elif data in ')':
                if punct_stack and punct_stack[-1] == '(':
                    punct_stack.pop()
                prefix = ''
            elif data in ']':
                if punct_stack and punct_stack[-1] == '[':
                    punct_stack.pop()
                prefix = ''
            elif data == '/…':
                prefix = ''
        
        elif tok.kind == Token.END_OF_SENTENCE:
            prefix = end_sentence
            if capitalize_opt:
                capitalize_next_word = True

        elif parts and parts[-1]:
            last_char = parts[-1][-1]
            if last_char == '«':
                prefix = '\xa0'
            elif punct_stack and last_char == punct_stack[-1]:
                prefix = ''
            elif punct_stack and last_char == '“':
                prefix = ''
            elif last_char not in '-/':
                prefix = ' '

        if parts:
            parts.append(prefix + data)
        else:
            # First word in sentence
            parts.append(data if data else '')
        if parts[-1] == '':
            parts.pop()
    
    if parts:
        ret = ''.join(parts)
        return ret
    return ''





def generate_raw_tokens(text_or_gen: Union[str, Iterable[str]]) -> Iterator[Token]:
    """ 
        Generate raw tokens by splitting strings on whitespaces

        <SPECIAL_TOKENS> will be generated at this stage as well
    """
    
    if isinstance(text_or_gen, str):
        if not text_or_gen:
            return
        text_or_gen = [text_or_gen]
    
    for sentence in text_or_gen:
        for s in sentence.split():
            if re.fullmatch(r"<[A-Z']+>", s):
                yield Token(s, Token.SPECIAL_TOKEN)
            else:
                yield Token(s, Token.RAW)


def generate_tokens_lookahead(token_stream: Iterator[Token]):
    try:
        token = next(token_stream)
    except StopIteration:
        return
    for next_token in token_stream:
        token.next = next_token
        yield token
        token = next_token
    token.next = None
    yield token



def generate_eos_tokens(token_stream: Iterator[Token]) -> Iterator[Token]:
    """
        Adds <END_OF_SENTENCE> tokens to token stream.

        Tokens must have the 'next' parameter defined by calling
        'generate_tokens_lookahead' on the token stream first.
    """

    subsentence_depth = 0
    in_double_quotes = False
    first_in_sentence = True

    for token in token_stream:
        if first_in_sentence and not token.kind == Token.PUNCTUATION:
            token.flags.add(Flag.FIRST_WORD)
            first_in_sentence = False
        
        if token.kind != Token.PUNCTUATION:
            yield token
            continue

        if Flag.OPENING_PUNCT in token.flags:
            subsentence_depth += 1
        elif Flag.CLOSING_PUNCT in token.flags:
            subsentence_depth -= 1
        elif token.data == '"':
            in_double_quotes = not in_double_quotes
            subsentence_depth += 1 if in_double_quotes else -1

        if token.norm:
            punct = token.norm[0]
        else:
            punct = token.data
        
        if punct in ".!?":
            yield token
            first_in_sentence = True
            if subsentence_depth == 0:
                yield Token('', Token.END_OF_SENTENCE)
        elif punct == '…' or re.fullmatch(r"\.\.+", punct):
            yield token
            first_in_sentence = True
            if subsentence_depth == 0 and token.next and is_capitalized(token.next.data):
                yield Token('', Token.END_OF_SENTENCE)
        else:
            yield token



def parse_punctuation(token_stream: Iterator[Token], **options: Any) -> Iterator[Token]:
    """
        Parse a stream of raw tokens to find punctuation marks

        Options:
            * norm_punct: Normalize punctuation marks
    
        TODO:
            * URLS, email addresses (anything with internal punctuation marks)
            * Fix: A... -> A.…
            * Fix: Y. -B. Piriou
    """

    # Normalize punctuation option
    norm_punct = options.pop('norm_punct', False)

    for tok in token_stream:
        if tok.kind == Token.RAW:
            data = tok.data
            subtokens = []
            post_subtokens = []

            while data:
                #print(f"'{data}'")

                # Check at the beggining of the token
                # Check for opening punctuation
                if data[0] in OPENING_PUNCT:
                    # punct_stack.append(data[0])
                    subtokens.append(Token(data[0], Token.PUNCTUATION, Flag.OPENING_PUNCT))
                    data = data[1:]
                    continue

                # Check for closing punctuation
                if data[0] in CLOSING_PUNCT:
                    # if punct_stack and data[0] == PUNCT_PAIRS[punct_stack[-1]]:
                    #     punct_stack.pop()
                    subtokens.append(Token(data[0], Token.PUNCTUATION, Flag.CLOSING_PUNCT))
                    data = data[1:]
                    continue
                
                # Check for ellipsis
                m = re.match(r"\.\.+", data)
                if m:
                    subtokens.append(Token(m.group(), Token.PUNCTUATION))
                    data = data[m.end():]
                    continue
                
                # Check for single punctuation mark
                if data[0] in PUNCTUATION:
                    subtokens.append(Token(data[0], Token.PUNCTUATION))
                    data = data[1:]
                    continue
                

                # Check for common abbreviations
                skip = False
                for abbr in abbreviations:
                    # pattern = abbr.replace('.', '\\.')
                    # m = re.fullmatch(pattern, data, re.IGNORECASE)
                    # if m:
                    if data == abbr:
                        t = Token(data, Token.ABBREVIATIION)
                        t.norm.append(abbreviations[abbr])
                        subtokens.append(t)
                        data = data[len(abbr):]
                        skip = True
                        break
                if skip:
                    continue

                m = re.match(r"h\.a(?=…)", data, re.IGNORECASE)
                if m:
                    t = Token(m.group(), Token.ABBREVIATIION)
                    t.norm.append("hag all")
                    subtokens.append(t)
                    subtokens.append(Token('…', Token.PUNCTUATION))
                    data = data[4:]
                    continue


                # Single initial or group of initials (i.e: I.E. or U.N.…)
                m = re.match(r"([A-Z]\.)+", data)
                if m:
                    subtokens.append(Token(m.group(), Token.ACRONYM))
                    data = data[m.end():]
                    continue
                

                # Check for trailing punctuation
                # Trailing ellipsis
                m = re.search(r"\.\.+$", data)
                if m:
                    post_subtokens.insert(0, Token(m.group(), Token.PUNCTUATION))
                    data = data[:m.start()]
                    continue

                # Other trailing punctuation
                m = re.match(r"\!(\!)*$", data)
                if m:
                    t = Token(m.group(), Token.PUNCTUATION)
                    t.norm.append('!')
                    post_subtokens.insert(0, t)
                    data = data[:m.start()]
                    continue
                
                m = re.match(r"\?(\?)*$", data)
                if m:
                    t = Token(m.group(), Token.PUNCTUATION)
                    t.norm.append('?')
                    post_subtokens.insert(0, t)
                    data = data[:m.start()]
                    continue
                
                if data[-1] in CLOSING_PUNCT:
                    post_subtokens.insert(0, Token(data[-1], Token.PUNCTUATION, Flag.CLOSING_PUNCT))
                    data = data[:-1]
                    continue

                if data[-1] in PUNCTUATION:
                    post_subtokens.insert(0, Token(data[-1], Token.PUNCTUATION))
                    data = data[:-1]
                    continue

                # Check for SI_UNITS here maybe

                # Parse the remainder
                m = re_extended_word.match(data)             # Doesn't match (covid-19)
                #m = re.match(r"[\w\-'’·]+", data)  # Breaks numbers with dots or commas
                # m = common_word.match(data)       # Doesn't match the final hyphen (labour- \ndouar )
                if m:
                    l = m.end() - m.start()
                    subtokens.append(Token(m.group()))
                    data = data[l:]
                    continue

                subtokens.append(Token(data))
                data = ''
            
            for t in subtokens + post_subtokens:
                if norm_punct:
                    if re.fullmatch(r"\.\.+", t.data):
                        t.norm.append('…')
                    elif t.data == '‚':   # dirty comma
                        t.norm.append(',')
                yield t
        
        else:
            yield tok



def parse_regular_words(token_stream: Iterator[Token], **options: Any) -> Iterator[Token]:
    """ It should be called after `parse_punctuation`
    
        TODO:
            * Brezhoneg/Galleg
            * miz Gouere.Laouen e oa
    """

    # Arg options

    for tok in token_stream:
        if tok.kind == Token.RAW:
            if tok.data in acronyms:
                tok.kind = Token.ACRONYM
            elif tok.data.isupper() and Flag.FIRST_WORD not in tok.flags:
                tok.kind = Token.ACRONYM
            elif is_word(tok.data):
                # Token is a simple and well formed word
                if is_proper_noun(tok.data):
                    tok.kind = Token.PROPER_NOUN
                else:
                    tok.kind = Token.WORD
                    if is_word_inclusive(tok.data):
                        tok.flags.add(Flag.INCLUSIVE)
            yield tok
        else:
            yield tok



def parse_numerals(token_stream: Iterator[Token]) -> Iterator[Token]:
    """ Look for various numeral forms: numbers, ordinals, units...
        It should be applied before `parse_regular_words` to accurately parse quantities

        TODO:
            * 1,20
            * €/miz
            * +40 %
            * d'ar Sul 10/10
            * ½
            * 02.98.00.00.00
            * 2003-2004
    """

    # prev_token = None
    num_concat = "" # buffer to contatenate numeral forms such as '12 000' -> '12000'
    for tok in token_stream:
        if tok.kind == Token.RAW:
            # r"[+-]?\d+(?:,\d+)"

            if tok.data.isdecimal():
                if not num_concat and len(tok.data) < 4:
                    num_concat += tok.data
                elif num_concat and len(tok.data) == 3:
                    num_concat += tok.data                    
                else:
                    # A full number
                    tok.kind = Token.NUMBER
            elif re.fullmatch(r"\d{1,3}(\.\d\d\d)+", tok.data):
                # Big number with dotted thousands (i.e: 12.000.000)
                tok.data = tok.data.replace('.', '')
                tok.kind = Token.NUMBER
            else:
                if is_roman_number(tok.data):
                    tok.kind = Token.ROMAN_NUMBER
                elif is_ordinal(tok.data):
                    tok.kind = Token.ORDINAL
                elif is_roman_ordinal(tok.data):
                    tok.kind = Token.ROMAN_ORDINAL
                elif is_time(tok.data):
                    # TODO: Check for token 'gm', 'g.m', 'GM'...
                    tok.kind = Token.TIME
                elif is_unit_number(tok.data):
                    # ex: "10m2"
                    number, unit = match_unit_number(tok.data).groups()
                    number = number.replace('.', '')
                    tok.kind = Token.QUANTITY
                    tok.data = f"{num_concat}{number}{unit}"
                    tok.number = num_concat + number
                    tok.unit = unit
                    if num_concat:
                        num_concat = ""
                elif tok.data in SI_UNITS:
                    if num_concat:
                        # ex: "10 s"
                        tok.kind = Token.QUANTITY
                        tok.number = num_concat
                        tok.unit = tok.data
                        tok.data = num_concat + tok.data
                        num_concat = ""
                    elif tok.data not in ('l', 'm', 't', 'g'):
                        tok.kind = Token.UNIT
                elif num_concat and is_noun(tok.data):
                    # ex: "32 bloaz"
                    tok.kind = Token.QUANTITY
                    tok.number = num_concat
                    tok.unit = tok.data
                    if is_word_inclusive(tok.data):
                        tok.flags.add(Flag.INCLUSIVE)
                    else:
                        if is_noun_f(tok.data):
                            tok.flags.add(Flag.FEMININE)
                        if is_noun_m(tok.data):
                            tok.flags.add(Flag.MASCULINE)
                    tok.data = f"{num_concat} {tok.data}"
                    num_concat = ""
                
                if num_concat:
                    yield Token(num_concat, Token.NUMBER)
                    num_concat = ""
        
        else:
            if num_concat:
                yield Token(num_concat, Token.NUMBER)
                num_concat = ""

        if not num_concat:
            yield tok
    
    if num_concat:
        yield(Token(num_concat, Token.NUMBER))



def correct_tokens(token_stream: Iterator[Token]) -> Iterator[Token]:
    """
        Correct words from `corrected_tokens.tsv` and `standard_tokens.tsv`.
        Should be applied before `parse_regular_words`
    """

    def get_susbitution(word: str) -> List[str]:
        lowered = word.lower()
        
        if lowered in corrected_tokens:
            substitutes = corrected_tokens[lowered]
        elif lowered in standard_tokens:
            substitutes = standard_tokens[lowered]
        else:
            return []
        
        # Keep capitalization
        i = 0
        while lowered[i] not in LETTERS: i += 1
        if word[i].isupper():
            first = capitalize(substitutes[0])
            return [first] + substitutes[1:]
        else:
            return substitutes

        


    for tok in token_stream:
        if tok.kind == Token.RAW:
            lowered = tok.data.lower()
            substitutes = get_susbitution(tok.data)
            if substitutes:
                # We must keep the prepended apostrophe (there could be a substitution rule for it)
                yield from [ Token(s, Token.RAW, Flag.CORRECTED) for s in substitutes ]
            elif lowered.startswith("'") and lowered[1:] not in ('n', 'm', 'z'):
                # Remove prepended apostrophies
                # Check if there is a susbstitution rule for the remaining word
                substitutes = get_susbitution(tok.data[1:])
                if substitutes:
                    yield from [ Token(s, Token.RAW, Flag.CORRECTED) for s in substitutes ]
                else:
                    # Pass the word without the apostrophe
                    tok.data = tok.data[1:]
                    yield tok
            else:
                yield tok
        else:
            yield tok


from typing import Iterator, Any

from .tokenizer import Token, tokenize, detokenize, split_sentences, split_sentences_old
from .normalizer import normalize, normalize_sentence
from .inverse_normalizer import inverse_normalize_sentence, inverse_normalize_timecoded
from .utils import (
    strip_punct, filter_out_chars, filter_in_chars, capitalize, pre_process,
    extract_parenthesis_content, sentence_stats,
)
from .definitions import PUNCTUATION, LETTERS, PUNCT_PAIRS, VALID_CHARS
from ..utils import read_file_drop_comments


def load_translation_dict(path: str) -> dict:
    translation_dict = dict()
    with open(path, 'r', encoding='utf-8') as f:
        for line in f.readlines():
            line = line.strip()
            if '\t' in line:
                key, val = line.split('\t')
                translation_dict[key] = val
            else:
                translation_dict[line] = ""
    return translation_dict



def reverse_translation_dict(path: str, newpath: str) -> None:
    """ Build a translation dictionary (tsv file) by reversing another translation dictionary
    """
    reversed = dict()
    for line in read_file_drop_comments(path):
        line = pre_process(line)
        if '\t' in line:
            key, val = line.split('\t')
            if ' ' in key or ' ' in val:
                print("no spaces allowed in translation dictionaries")
                continue
            if val in reversed:
                reversed[val] += ", {}".format(key)
            else:
                reversed[val] = key
    with open(newpath, 'w', encoding='utf-8') as f:
        for k in sorted(reversed):
            f.write(f"{k}\t{reversed[k]}\n")



def correct_sentence(sentence: str) -> str:
    return detokenize(tokenize(sentence, autocorrect=True))



def translate_tokens(token_stream: Iterator[Token], tra_dict: dict, **options: Any) -> Iterator[Token]:
    """ Substitute tokens according to a given dictionary
        
        Keys with uppercase letters will be case-sentitive
        Keys with lowercase letters only will be case-insensitive
        Translate will operate according to the first match in the dictionary

        Key/value pairs of the translation dictionary can contain
        the '*' character to match any character
        Ex: "*a" : "*añ"    -> will change suffixes of words ending with 'a'
    """

    for tok in token_stream:
        for key, val in tra_dict.items():
            if key.startswith('*'):
                if key.endswith('*'):
                    # Match chars in the middle of words
                    expr = key.strip('*')
                    if expr in tok.data:
                        tok.data.replace(expr, val.strip('*'))
                else:
                    # Match at the end
                    expr = key.lstrip('*')
                    if tok.data.endswith(expr):
                        tok.data = tok.data[:-len(expr)] + val.lstrip('*')
            elif key.endswith('*'):
                # Match at the beginning
                expr = key.rstrip('*')
                if tok.data.startswith(expr):
                    tok.data = val.rstrip('*') + tok.data[len(expr):]
            elif key.islower():
                if key == tok.data.lower():
                    if tok.data.istitle():
                        tok.data = val.capitalize()
                    elif tok.data.isupper():
                        tok.data = val.upper()
                    else:
                        tok.data = val
                    break
            else:
                if key == tok.data:
                    tok.data = tra_dict[tok.data]
                    break
        yield tok


def count_words(sentence: str) -> int:
    """ Return number of regular words in sentence """
    n = 0
    for t in tokenize(sentence, norm_punct=True, autocorrect=True):
        if t.kind == Token.WORD:
            n += 1
    return n


def is_full_sentence(sentence: str) -> bool:
    """Check if sentence is a complete sentence, punctuation-wise"""
    return sentence[0].isupper() and sentence[-1] in ".!?…"

def is_sentence_start_open(sentence: str) -> bool:
    return (
        sentence[0].islower()
        or (sentence[0].isupper() and (sentence[1].isupper() or sentence[1].isdigit()))  # Acronym
        or sentence[0] in "'’"
        or sentence[0].isdigit()
    )
	
def is_sentence_end_open(sentence: str) -> bool:
    return (
        sentence[-1].islower()
        or sentence[-1] in "…'’,»\""
        or sentence[-1].isdigit()
    )

def is_sentence_punct_paired(sentence: str) -> bool:
    if sentence.count('"') % 2 != 0:
        return False
    for punct in PUNCT_PAIRS:
        if sentence.count(punct) != sentence.count(PUNCT_PAIRS[punct]):
            return False
    return True


from typing import Union, List
import os

from colorama import Fore



def green(s:str) -> str:
    return Fore.GREEN + s + Fore.RESET

def yellow(s:str) -> str:
    return Fore.YELLOW + s + Fore.RESET

def red(s:str) -> str:
    return Fore.RED + s + Fore.RESET



def sec2hms(seconds):
    """ Return a string of hours, minutes, seconds from a given number of seconds """
    minutes, seconds = divmod(round(seconds), 60)
    hours, minutes = divmod(minutes, 60)
    return f"{hours}h {minutes}' {seconds}''"



def list_files_with_extension(ext: Union[str, tuple, list], rep, recursive=True) -> List[str]:
    """
        Recursively list all files of the given extension(s) in a folder

        Parameters
        ----------
            ext : str|list
                file extension, including the extension separator (dot)
    """
    extensions = (ext) if isinstance(ext, str) else ext
    file_list = []
    if os.path.isdir(rep):
        for filename in os.listdir(rep):
            filename = os.path.join(rep, filename)
            if os.path.isdir(filename) and recursive:
                file_list.extend(list_files_with_extension(extensions, filename))
            else:
                file_ext = os.path.splitext(filename)[1]
                if file_ext and file_ext in extensions:
                    file_list.append(filename)
    return file_list



def read_file_drop_comments(path: str) -> List[str]:
    lines = []
    with open(path, 'r', encoding='utf-8') as f:
        for l in f.readlines():
            l = l.strip()
            if l and not l.startswith('#'):
                lines.append(l)
    return lines


from typing import Tuple, List, Dict
import sys
import os.path
import re
from math import floor, ceil

from hashlib import md5
from uuid import uuid4
from colorama import Fore

# For eaf (Elan) file conversion
from xml.dom import minidom
import datetime, pytz

from ..utils import read_file_drop_comments
from ..text import (
    pre_process, normalize_sentence, filter_out_chars,
    split_sentences, tokenize, detokenize, normalize, Token,
    PUNCTUATION,
    VALID_CHARS
)
from ..audio import convert_to_mp3, find_associated_audiofile
from ..version import VERSION


datafile_header = \
"""{source: }
{source-audio: }
{author: }
{licence: }
{tags: }\n\n
"""


# Special tokens, found in transcriptions
special_tokens = {
    "<UNK>": "SPN",
    "<SPOKEN_NOISE>": "SPN",
    "<NTT>": "SPN",
    "<C'HOARZH>": "LAU",
    "<HUM>": "SPN",
    "<PASAAT>": "SPN",
    "<SONEREZH>": "NSN",
    "<FRONAL>": "SPN",
}



Segment = Tuple[int, int]

def load_segments_data(segfile: str) -> List[Segment]:
    """ Load audio segments delimiters from a `.seg` file
        Return a list of segments
    """

    segments = []
    with open(segfile, 'r') as f:
        for l in f.readlines():
            l = l.strip()
            if not l or l.startswith('#'):
                continue
            t = l.split()
            start = int(t[0]) / 1000
            stop = int(t[1]) / 1000
            segments.append((start, stop))
            
    return segments



def load_text_data(filename) -> List[Tuple[str, Dict]]:
    """ 
        Return list of sentences with metadata.
        Metadata dictionaries will always have, at least, the "speaker" and "gender" keys.

        Return
        ------
            list of tuple (text sentences, metadata)
    """
    utterances = []
    current_speaker = 'unknown'
    current_gender = 'unknown'
    no_lm = False
    for l in read_file_drop_comments(filename):
        # Extract speaker id and other metadata
        l, metadata = extract_metadata(l)
        if "speaker" in metadata:
            current_speaker = metadata["speaker"]
        else:
            metadata["speaker"] = current_speaker
        
        if "gender" in metadata:
            current_gender = metadata["gender"]
        else:
            metadata["gender"] = current_gender
        
        if "parser" in metadata:
            if "no-lm" in metadata["parser"]: no_lm = True
            elif "add-lm" in metadata["parser"]: no_lm = False
        else:
            if no_lm:
                metadata["parser"] = ["no-lm"]
        if l:
            utterances.append((l, metadata))
    return utterances



def get_text_header(filename) -> Dict:
    metadata = dict()
    for l in read_file_drop_comments(filename):
        l, md = extract_metadata(l)
        metadata.update(md)
        if l:
            # Stop a first sentence
            break
    return metadata



def format_timecode(timecode):
    if isinstance(timecode, int):
        return str(timecode)
    return "{:.3f}".format(timecode).rstrip('0').rstrip('.')



def create_ali_file(sentences, segments, **kwargs) -> str:
    """
        Common header metadata:
            audio-path
            source
            audio-source
            author
            transcription: name of transcribers
            licence
            tags: list of tags separated by commas
    """
    data = []

    if value := kwargs.pop("audio-path", None):
        data.append(f"{{audio-path: {value}}}")
    if value := kwargs.pop("audio_path", None):
        data.append(f"{{audio-path: {value}}}")
    for key, value in sorted(kwargs.items()):
        key = key.replace('_', '-')
        data.append(f"{{{key}: {value}}}")
    data.append("")

    for sentence, segment in zip(sentences, segments):
        start = format_timecode(segment[0])
        end = format_timecode(segment[1])
        data.append(f"{sentence.strip()} {{start: {start}; end: {end}}}")
    
    return '\n'.join(data)



def load_ali_file(filepath) -> Dict:
    """
        Return a dictionary containing the data from an `ali` file :
            audio_path: str,
            sentences: list,
            raw_sentences: list,
            segments: list,
            metadata: list,

    """

    audio_path = None
    header = dict()
    sentences = []       # Text without metadata
    raw_sentences = []   # Text with metadata
    segments = []        # Segments in milliseconds
    metadatas = []

    with open(filepath, 'r', encoding='utf-8') as f:
        # Find associated audio file in metadata
        current_speaker = 'unknown'
        current_gender = 'unknown'
        no_lm = False

        for line in f.readlines():
            line = line.strip()
            if line.startswith('#'):
                continue

            text, metadata = extract_metadata(line)
            if "speaker" in metadata:
                current_speaker = metadata["speaker"]
            else:
                metadata["speaker"] = current_speaker
            
            if "gender" in metadata:
                current_gender = metadata["gender"]
            else:
                metadata["gender"] = current_gender
            
            if "parser" in metadata:
                if "no-lm" in metadata["parser"]: no_lm = True
                elif "add-lm" in metadata["parser"]: no_lm = False
            else:
                if no_lm:
                    metadata["parser"] = ["no-lm"]

            # match = re.search(r"{\s*start\s*:\s*([0-9\.]+)\s*;\s*end\s*:\s*([0-9\.]+)\s*}", line)
            # if match:
            if "start" in metadata and "end" in metadata:
                segments.append([metadata["start"], metadata["end"]])
                sentences.append(text.strip())
                raw_sentences.append(line)
                metadatas.append(metadata)
            elif metadata:
                metadata.pop("speaker", None)
                metadata.pop("gender", None)
                metadata.pop("accent", None)
                header.update(metadata)

            if not audio_path and "audio-path" in metadata:
                dir = os.path.split(filepath)[0]
                audio_path = os.path.join(dir, metadata["audio-path"])
                audio_path = os.path.normpath(audio_path)
        
        # Try to find an associated audiofile if it was not explicitely set in metadata
        if not audio_path:
            audio_path = find_associated_audiofile(filepath, silent=True)
    
    return {
        "audio_path": audio_path,
        "header": header,
        "sentences": sentences,
        "raw_sentences": raw_sentences,
        "segments": segments,
        "metadata": metadatas,
    }



def parse_dataset(file_or_dir, args):
    if (file_or_dir.endswith(".split")
        or file_or_dir.endswith(".seg")
        or file_or_dir.lower().endswith('.ali')
    ):   # Single data item
        return parse_data_file(file_or_dir, args)

    elif os.path.isdir(file_or_dir):
        data = {
            "path": file_or_dir,
            "wavscp": [],   # Recording id to wave filenames
            "utt2spk": [],      # Utterance id to speakers id
            "segments": [],     # Time segments
            "text": [],         # Utterances it to text
            "speakers": set(),  # Speakers names
            "lexicon": set(),   # Word dictionary
            "corpus": set(),    # Sentences for LM corpus
            "audio_length": {'m': 0, 'f': 0, 'u': 0},    # Audio length for each gender
            "subdir_audiolen": {}   # Size (total audio length) for every sub-folders
            }
        
        for filename in sorted(os.listdir(file_or_dir)):
            if filename.startswith('.'):
                # Skip hidden folders
                continue
            if (os.path.isdir(os.path.join(file_or_dir, filename))
                or filename.endswith(".split")
                or filename.endswith(".seg")
                or filename.lower().endswith(".ali")
            ):
                item_data = parse_dataset(os.path.join(file_or_dir, filename), args)
                data["wavscp"].extend(item_data["wavscp"])
                data["utt2spk"].extend(item_data["utt2spk"])
                data["segments"].extend(item_data["segments"])
                data["text"].extend(item_data["text"])
                data["speakers"].update(item_data["speakers"])
                data["lexicon"].update(item_data["lexicon"])
                data["corpus"].update(item_data["corpus"])
                data["subdir_audiolen"][filename] = \
                    item_data["audio_length"]['f'] + \
                    item_data["audio_length"]['m'] + \
                    item_data["audio_length"]['u']
                for k, dur in item_data["audio_length"].items():
                    if k in data["audio_length"]:
                        data["audio_length"][k] += dur
                    else:
                        data["audio_length"][k] = dur

        return data
    else:
        print("File argument must be a split file or a directory")
        return
    


valid_chars = set(VALID_CHARS)
speakers_gender = {"unknown": 'u'}

def parse_data_file(filepath, args):
    print(Fore.GREEN + f" * {filepath}" + Fore.RESET, end=' ', flush=True)
    
    seg_ext = os.path.splitext(filepath)[1] # Could be '.split' or '.seg'
    audio_path = ""

    if seg_ext == ".ali":
        aligned_data = load_ali_file(filepath)
        segments = aligned_data["segments"]
        sentences_and_metadata = list(zip(aligned_data["sentences"], aligned_data["metadata"]))
        audio_path = aligned_data["audio_path"]
    else: # .seg, .split
        text_filename = filepath.replace(seg_ext, '.txt')
        assert os.path.exists(text_filename), f"ERROR: no text file found for {filepath}"
        segments = load_segments_data(filepath)
        sentences_and_metadata = load_text_data(text_filename)
    assert len(sentences_and_metadata) == len(segments), \
        f"number of utterances in text file ({len(data['text'])}) doesn't match number of segments in split file ({len(segments)})"

    
    # Look for accompanying audio file
    if not audio_path:
        audio_path = os.path.abspath(filepath.replace(seg_ext, '.wav'))
        if not os.path.exists(audio_path):
            audio_path = os.path.abspath(filepath.replace(seg_ext, '.mp3'))
    assert os.path.exists(audio_path), f"ERROR: no audio file found for {filepath}"
    
    recording_id = md5(audio_path.encode("utf8")).hexdigest()

    # Use a single random speaker id per file for unknown speakers
    unk_speaker_id = str(uuid4()).replace('-', '')
    

    data = {
        "wavscp": [],       # Recording id to wave filenames
        "utt2spk": [],      # Utterance id to speakers id
        "segments": [],     # Time segments
        "text": [],         # Utterances id to text
        "speakers": set(),  # Speakers names
        "lexicon": set(),   # Word dictionary
        "corpus": set(),    # Sentences for LM corpus
        "audio_length": {'m': 0, 'f': 0, 'u': 0},    # Audio length for each gender
        }
    

    # if not args.split_audio:
    data["wavscp"].append((recording_id, audio_path))

    for (sentence, metadata), (start, end) in zip(sentences_and_metadata, segments):
        if end - start < args.utt_min_len:
            # Skip short utterances
            print(Fore.YELLOW + "dropped (too short): " + Fore.RESET + sentence, file=sys.stderr)
            continue
            
        speaker_id = metadata["speaker"]

        if speaker_id == "unknown":
            speaker_id = unk_speaker_id
        else:
            if args.hash_id:
                speaker_id = md5(speaker_id.encode('utf-8')).hexdigest()
            data["speakers"].add(speaker_id)
        
        utterance_id = f"{speaker_id}-{recording_id}-{floor(100*start):0>7}_{ceil(100*end):0>7}"
        
        if speaker_id not in speakers_gender:
            # speakers_gender is a global variable
            speakers_gender[speaker_id] = metadata["gender"]
        
        cleaned_sentence = pre_process(sentence)
        tokens = list(tokenize(cleaned_sentence, autocorrect=True, norm_punct=False))
        sent = detokenize(normalize(tokens, norm_case=True), normalize=True, capitalize=False)
        
        sent = sent.replace('\xa0', ' ') # Non-breakable spaces
        sent = sent.replace('-', ' ').replace('/', ' ')
        sent = filter_out_chars(sent, PUNCTUATION + '*')
        if not sent:
            continue
        sent = ' '.join(sent.split())
        
        # Filter out utterances with numbers or foreign chars (not counting acronyms)
        sent_no_acronyms = detokenize(
            normalize(
                filter(lambda t: not t.kind == Token.ACRONYM, tokens),
                norm_case=True
            ), normalize=True
        )
        sent_no_acronyms = sent_no_acronyms.replace('\xa0', ' ').replace('*', '')
        chars = set(sent_no_acronyms)
        if not chars.issubset(valid_chars):
            print('\n' + Fore.YELLOW
                + f"dropped (foreign chars '{chars.difference(valid_chars)}'): "
                + Fore.RESET + sentence, end='', file=sys.stderr)
            continue
        
        data["text"].append((utterance_id, sent))
        data["utt2spk"].append((utterance_id, speaker_id))
        data["segments"].append((utterance_id, recording_id, start, end))


        # Keeping track of gender representation
        if metadata["gender"] == 'm':
            data["audio_length"]['m'] += end - start
        elif metadata["gender"] == 'f':
            data["audio_length"]['f'] += end - start
        else:
            data["audio_length"]['u'] += end - start


        # Add words to lexicon
        for word in sent.split():
            # Remove black-listed words (those beggining with '*')
            if word.startswith('*'):
                pass
            elif word in special_tokens:
                pass
            elif word == "'":
                pass
            else: data["lexicon"].add(word)


        # Add sentence to language model corpus
        add_to_text_corpus = True
        if "parser" in metadata:
            if "no-lm" in metadata["parser"]:
                add_to_text_corpus = False
            elif "add-lm" in metadata["parser"]:
                add_to_text_corpus = True
        
        if add_to_text_corpus and not args.no_lm:
            for sub in split_sentences(cleaned_sentence):
                sent = normalize_sentence(sub, autocorrect=True, norm_case=True)
                sent = sent.replace('-', ' ').replace('/', ' ')
                sent = sent.replace('\xa0', ' ')
                sent = filter_out_chars(sent, PUNCTUATION)
                if not sent.strip():
                    continue

                n_stared = sent.count('*')
                tokens = sent.split()
                # Ignore if to many black-listed words in sentence
                if n_stared / len(tokens) > 0.2:
                    if args.verbose:
                        print(Fore.YELLOW + "LM exclude:" + Fore.RESET, sent, end='')
                    continue
                # Remove starred words
                tokens = [tok for tok in tokens if not tok.startswith('*')]
                # Ignore if sentence is too short
                if len(tokens) < args.lm_min_token:
                    if args.verbose:
                        print(Fore.YELLOW + "LM exclude:" + Fore.RESET, sent, end='')
                    continue
                data["corpus"].add(' '.join(tokens))
     
    # status = Fore.GREEN + f" * {filepath[:-6]}" + Fore.RESET
    if data["audio_length"]['u'] > 0:
        print(' ' + Fore.RED + "unknown speaker(s)" + Fore.RESET, end='')
    print()
    return data



def create_eaf(segments, sentences, audiofile, type="wav"):
    """ Export to eaf (Elan) file """

    record_id = os.path.splitext(os.path.abspath(audiofile))[0]
    if type == "mp3":
        mp3_file = os.path.extsep.join((record_id, 'mp3'))
        if not os.path.exists(mp3_file):
            convert_to_mp3(audiofile, mp3_file)
        audiofile = mp3_file

    doc = minidom.Document()

    root = doc.createElement('ANNOTATION_DOCUMENT')
    root.setAttribute('AUTHOR', f'Anaouder {VERSION} (Gweltaz DG)')
    root.setAttribute('DATE', datetime.datetime.now(pytz.timezone('Europe/Paris')).isoformat(timespec='seconds'))
    root.setAttribute('FORMAT', '3.0')
    root.setAttribute('VERSION', '3.0')
    root.setAttribute('xmlns:xsi', 'http://www.w3.org/2001/XMLSchema-instance')
    root.setAttribute('xsi:noNamespaceSchemaLocation', 'http://www.mpi.nl/tools/elan/EAFv3.0.xsd')
    doc.appendChild(root)

    header = doc.createElement('HEADER')
    header.setAttribute('MEDIA_FILE', '')
    header.setAttribute('TIME_UNITS', 'milliseconds')
    root.appendChild(header)

    media_descriptor = doc.createElement('MEDIA_DESCRIPTOR')
    media_descriptor.setAttribute('MEDIA_URL', 'file://' + os.path.abspath(audiofile))
    if type == "mp3":
        media_descriptor.setAttribute('MIME_TYPE', 'audio/mpeg')
    else:
        media_descriptor.setAttribute('MIME_TYPE', 'audio/x-wav')
    media_descriptor.setAttribute('RELATIVE_MEDIA_URL', './' + os.path.basename(audiofile))
    header.appendChild(media_descriptor)

    time_order = doc.createElement('TIME_ORDER')
    last_t = 0
    for i, (s, e) in enumerate(segments):
        if s < last_t:
            s = last_t
        last_t = s
        time_slot = doc.createElement('TIME_SLOT')
        time_slot.setAttribute('TIME_SLOT_ID', f'ts{2*i+1}')
        time_slot.setAttribute('TIME_VALUE', str(s))
        time_order.appendChild(time_slot)
        time_slot = doc.createElement('TIME_SLOT')
        time_slot.setAttribute('TIME_SLOT_ID', f'ts{2*i+2}')
        time_slot.setAttribute('TIME_VALUE', str(e))
        time_order.appendChild(time_slot)
    root.appendChild(time_order)

    tier_trans = doc.createElement('TIER')
    tier_trans.setAttribute('LINGUISTIC_TYPE_REF', 'transcript')
    tier_trans.setAttribute('TIER_ID', 'Transcription')

    for i, sentence in enumerate(sentences):
        annotation = doc.createElement('ANNOTATION')
        alignable_annotation = doc.createElement('ALIGNABLE_ANNOTATION')
        alignable_annotation.setAttribute('ANNOTATION_ID', f'a{i+1}')
        alignable_annotation.setAttribute('TIME_SLOT_REF1', f'ts{2*i+1}')
        alignable_annotation.setAttribute('TIME_SLOT_REF2', f'ts{2*i+2}')
        annotation_value = doc.createElement('ANNOTATION_VALUE')
        #text = doc.createTextNode(get_cleaned_sentence(sentence, rm_bl=True, keep_dash=True, keep_punct=True)[0])
        text = doc.createTextNode(sentence.replace('*', ''))
        annotation_value.appendChild(text)
        alignable_annotation.appendChild(annotation_value)
        annotation.appendChild(alignable_annotation)
        tier_trans.appendChild(annotation)
    root.appendChild(tier_trans)

    linguistic_type = doc.createElement('LINGUISTIC_TYPE')
    linguistic_type.setAttribute('GRAPHIC_REFERENCES', 'false')
    linguistic_type.setAttribute('LINGUISTIC_TYPE_ID', 'transcript')
    linguistic_type.setAttribute('TIME_ALIGNABLE', 'true')
    root.appendChild(linguistic_type)

    language = doc.createElement('LANGUAGE')
    language.setAttribute("LANG_ID", "bre")
    language.setAttribute("LANG_LABEL", "Breton (bre)")
    root.appendChild(language)

    constraint_list = [
        ("Time_Subdivision", "Time subdivision of parent annotation's time interval, no time gaps allowed within this interval"),
        ("Symbolic_Subdivision", "Symbolic subdivision of a parent annotation. Annotations refering to the same parent are ordered"),
        ("Symbolic_Association", "1-1 association with a parent annotation"),
        ("Included_In", "Time alignable annotations within the parent annotation's time interval, gaps are allowed")
    ]
    for stereotype, description in constraint_list:
        constraint = doc.createElement('CONSTRAINT')
        constraint.setAttribute('DESCRIPTION', description)
        constraint.setAttribute('STEREOTYPE', stereotype)
        root.appendChild(constraint)

    xml_str = doc.toprettyxml(indent ="\t", encoding="UTF-8")

    return xml_str.decode("utf-8")



def convert_from_eaf(eaf_filename):
    """ Write a split file and a text file from an eaf file """
    
    abs_path = os.path.abspath(eaf_filename)
    rep, eaf_filename = os.path.split(abs_path)
    
    print(rep, eaf_filename)
    print(abs_path)
    doc = minidom.parse(abs_path)
    root = doc.firstChild

    segments = []

    def getText(nodelist):
        rc = []
        for node in nodelist:
            if node.nodeType == node.TEXT_NODE:
                rc.append(node.data)
        return ''.join(rc)

    header = root.getElementsByTagName("HEADER")[0]
    md = header.getElementsByTagName("MEDIA_DESCRIPTOR")[0]
    wav_rel_path = md.getAttribute("RELATIVE_MEDIA_URL")

    wav_filename = os.path.normpath(os.path.join(rep, wav_rel_path))
    record_id = wav_filename.split(os.path.extsep)[0]
    text_filename = os.path.extsep.join((record_id, 'txt'))
    split_filename = os.path.extsep.join((record_id, 'split'))
    
    if os.path.exists(split_filename):
        print("Split file already exists.")
        while True:
            r = input("Replace (y/n)? ")
            if r.startswith('n'):
                print("Aborting...")
                return
            elif r.startswith('y'):
                break
    if not os.path.exists(wav_filename):
        print(f"Couldn't find '{wav_filename}'. Aborting...")
        return

    print("rep", rep)
    print("path", text_filename)

    time_order = root.getElementsByTagName("TIME_ORDER")[0]
    time_slots = time_order.getElementsByTagName("TIME_SLOT")
    time_slot_dict = {}
    for ts in time_slots:
        time_slot_dict[ts.getAttribute("TIME_SLOT_ID")] = int(ts.getAttribute("TIME_VALUE"))

    tiers = root.getElementsByTagName("TIER")
    for tier in tiers:
        if tier.getAttribute("TIER_ID").lower() in ("transcription", "default") :
            annotations = tier.getElementsByTagName("ANNOTATION")
            for annotation in annotations:
                aa = annotation.getElementsByTagName("ALIGNABLE_ANNOTATION")[0]
                ts1 = aa.getAttribute("TIME_SLOT_REF1")
                ts2 = aa.getAttribute("TIME_SLOT_REF2")
                time_seg = (time_slot_dict[ts1], time_slot_dict[ts2])
                text = getText(aa.getElementsByTagName("ANNOTATION_VALUE")[0].childNodes)
                segments.append((time_seg, text))
                #print(f"SEG: {time_seg} {text}")

    with open(text_filename, 'w', encoding='utf-8') as f:
        f.write('#\n' * 4 + '\n' * 6)
        for _, sentence in segments:
            f.write(sentence + '\n')
    with open(split_filename, 'w', encoding='utf-8') as f:
        for (s, e), _ in segments:
            f.write(f"{s} {e}\n")




##############################  METADATA  ##############################

METADATA_PATTERN = re.compile(r'{\s*(.+?)\s*}') # Capture content
METADATA_UNIT_PATTERN = re.compile(r"([\w\s:,_'’/=\.\-\?\&]+)") # Capture content except ';'
SPEAKER_NAME_PATTERN = re.compile(r"(?:(?:spk|speaker)\s*:\s*)?([\w '_-]+?)")
SPEAKER_ID_PATTERN_DEPR = re.compile(r'([-\'\w]+):*([mf])*')
KEYVAL_PATTERN = re.compile(r"([\w_'-]+)\s*:\s*([\w ,_'’.:/-]+?)\s*")

_VALID_PARAMS = {
    "source",
    "source-audio", "audio-source",
    "audio-path",
    "author", "authors",
    "licence",
    "modifications",
    "transcription",
    "tags",
    "parser",
    "speaker", "spk",
    "gender",
    "lang",
    "accent",
    "start", "end",
#    "phon",
}


def extract_metadata(sentence: str) -> Tuple[str, dict]:
    """ Returns the sentence stripped of its metadata (if any)
        and a dictionary of metadata
        Keeps unknown word markers '{?}'
    """
    metadata = dict()
    remove_ranges = []

    for match in METADATA_PATTERN.finditer(sentence):
        start, end = match.span()
        content = match.group(1).strip()
        if content == '?':       # Unknown words {?}
            if "unknown" not in metadata: metadata["unknown"] = []
            sub = sentence[:end]
            metadata["unknown"].append(len(sub.split())-1) # word number
        else:
            remove_ranges.append((start, end))
            metadata_units = content.split(';')
            for unit in metadata_units:
                unit = unit.strip()
                if ':' in unit:
                    # Key-value pair
                    key, val = unit.split(':', maxsplit=1)
                    key = key.strip()
                    val = val.strip()

                    if key in _VALID_PARAMS:
                        if key in ("speaker", "spk"):
                            key = "speaker"
                            if not val.isupper():
                                val = val.replace(' ', '_').lower()
                        if key in ("tags", "author", "accent"):
                            val = [v.strip().replace(' ', '_') for v in val.split(',') if v.strip()]
                        if key in ("start", "end"):
                            val = float(val)
                        metadata[key] = val
                    else:
                        speaker_name_depr = SPEAKER_ID_PATTERN_DEPR.fullmatch(unit)
                        if speaker_name_depr:
                            print(Fore.RED + f"Deprecated metadata: {unit}" + Fore.RESET)
                            metadata["speaker"] = speaker_name_depr.group(1)
                            if speaker_name_depr.group(2) in 'fm':
                                metadata["gender"] = speaker_name_depr.group(2)
                        else:
                            print(Fore.RED + f"Wrong metadata: {unit}" + Fore.RESET)
                else:
                    # A simplified speaker name
                    if not unit.isupper():
                        # Keep all-caps names (Acronyms)
                        unit = unit.replace(' ', '_').lower()
                    metadata["speaker"] = unit

    nchars_removed = 0
    for start, end in remove_ranges:
        sentence = sentence[:start-nchars_removed] + sentence[end-nchars_removed:]
        nchars_removed += end-start
    
    return sentence.strip(), metadata



from typing import List, Iterator, Any, Tuple
import re
from .definitions import LETTERS, PUNCTUATION



def strip_punct(word: str) -> str:
    """ Strip punctuation left and right of a word """

    while word and word[0] in PUNCTUATION:
        word = word[1:]
    while word and word[-1] in PUNCTUATION:
        word = word[:-1]
    return word



def filter_out_chars(text: str, chars: str) -> str:
    """ Remove given characters from a string """

    filtered_text = ""
    for l in text:
        if not l in chars: filtered_text += l
    return filtered_text



def filter_in_chars(text: str, allowed_chars: str) -> str:
    """ Remove all characters that are not in allowed list """

    filtered_text = []
    for c in text:
        if c in allowed_chars: filtered_text.append(c)
    return ''.join(filtered_text)



def is_capitalized(word: str) -> bool:
    return word.istitle()



def capitalize(word: str) -> str:
    """
        Capitalize a single word or the first word of a given sentence
    
        TODO:
            * Capitalize compound names, i.e "marie-thérèse" -> "Marie-Thérèse"
        
        Should we capitalize the whole "C'H" character ? For now we don't.
    """

    # Find first letter
    lowered = word.lower()
    i = 0
    while i < len(lowered) and lowered[i] not in LETTERS: i += 1
    if i >= len(word):
        return word
    
    capitalized = word[:i] + word[i].upper() + word[i+1:]
    return capitalized



_PATTERN_BETWEEN_PARENTHESIS = re.compile(r"\((.+?)\)")

def extract_parenthesis_content(txt: str) -> Tuple[str, str]:
    extracted = []
    remaining = txt
    match = re.search(_PATTERN_BETWEEN_PARENTHESIS, remaining)
    while match:
        extracted.append(match.group(1))
        start, end = match.span()
        remaining = remaining[:start] + remaining[end:]
        match = re.search(_PATTERN_BETWEEN_PARENTHESIS, remaining)
    return remaining, extracted



def pre_process(text: str) -> str:
    """ Correct ambiguous quote characters, tilde and such """

    skrab = '\''
    # text = text.replace("c'h", f"c{skrab}h")
    # text = text.replace("C'h", f"C{skrab}h")
    # text = text.replace("C'H", f"C{skrab}H")
    text = text.replace('‘', skrab)
    text = text.replace('’', "'")
    text = text.replace('ʼ', skrab)
    text = text.replace(',', ',')
    text = text.replace('˜', '') # Found instead of non-breakable spaces when parsing Ya! pdfs
    text = text.replace('Š', '') # Found instead of long dashes when parsing Ya! pdfs
    text = text.replace('ñ', 'ñ') # A sneaky n-tilde (found in Ya! webpages)
    text = text.replace('ň', 'ñ')
    text = text.replace('ñ̃', 'ñ') # Found in brezhoweb subtitles
    text = text.replace('ù', 'ù') # Another sneaky one (found in Ya! webpages)
    text = text.replace('ê', 'ê') # Found in brezhoweb subtitles
    text = text.replace('û', 'û') # Found in brezhoweb subtitles
    text = text.replace('ı', 'i') # Found in #Brezhoneg newspaper
    text = text.replace('ã', 'a')
    text = text.replace('ö', 'o')
    text = text.replace('á', 'a')
    return text


def sentence_stats(sentence: str) -> dict:
    """ Get statistics about a text """
    
    letter = 0
    decimal = 0
    upper = 0
    punct = 0
    blank = 0
    other = 0
    
    for c in sentence:
        if c.lower() in LETTERS or c in "'-":
            letter += 1
            if c.isupper():
                upper += 1
        elif c.isdecimal():
            decimal += 1
        elif c.isspace():
            blank += 1
        elif c in PUNCTUATION:
            punct += 1
        else:
            other += 1
        
        sentence = filter_out_chars(sentence, PUNCTUATION)
    
    return {"letter": letter, "decimal": decimal, "upper": upper, "punct": punct,
            "blank": blank, "other": other, "words": len(sentence.split())}



from typing import List

import os
import sys
import platform
import json

import ssl
import certifi
import urllib
import zipfile
from tqdm import tqdm

from vosk import Model, SetLogLevel



MODEL_LIST_URL = "https://raw.githubusercontent.com/gweltou/patromou/refs/heads/main/model_list.json"
_certifi_context = ssl.create_default_context(cafile=certifi.where())

_model_list = None
_loaded_model = None
_loaded_model_name = ""



def _get_model_directory() -> str:
    if platform.system() in ("Linux", "Darwin"):
        default = os.path.join(os.path.expanduser("~"), ".cache")
    elif platform.system() == "Windows":
        default = os.getenv("LOCALAPPDATA")
    else:
        raise OSError("Unsupported operating system")
    model_dir = os.path.join(os.getenv("XDG_CACHE_HOME", default), "anaouder", "models")
    
    if not os.path.exists(model_dir):
        os.makedirs(model_dir, exist_ok=True)
    
    return model_dir

_model_root = _get_model_directory()



def _get_model_list():
    global _model_list
    if _model_list:
        return _model_list

    cached_path = os.path.join(_model_root, "model_list.json")

    try:
        _model_list = json.load(urllib.request.urlopen(MODEL_LIST_URL, timeout=4, context=_certifi_context))
    except (urllib.error.URLError, urllib.error.HTTPError) as error:
        print(f"Cannot access online model list: {error}", file=sys.stderr)
        print(f"Reverting to cached model list", file=sys.stderr)
        _model_list = json.load(open(cached_path, 'r'))
    else:
        # Save a cached copy of `model_list.json`
        cached_model_list = []
        if os.path.exists(cached_path):
            try:
                cached_model_list = json.load(open(cached_path, 'r'))
            except json.JSONDecodeError:
                pass
        if _model_list != cached_model_list:
            json.dump(_model_list, open(cached_path, 'w'))
    return _model_list



def get_all_models() -> List[str]:
    """
    Returns the names of available models online
    """
    models = _get_model_list()
    return [ model['name'] for model in models ]



def get_local_models() -> List[str]:
    """
    Returns the names of locally available models
    """
    valid_models = []
    for f in os.listdir(_model_root):
        path = os.path.join(_model_root, f)
        if _is_valid_vosk_model(path):
            valid_models.append(f)
    return valid_models



def get_latest_model(type=None) -> str:
    models = sorted(_get_model_list(), key=lambda m: m["version"], reverse=True)
    if type:
        models = list(filter(lambda m: m["type"]==type, models))
    return models[0]["name"]



_MODEL_ALIASES = {
    "vosk-br-0.6": "vosk-model-br-0.6",
    "vosk6": "vosk-model-br-0.6",
    "vosk-br-0.7": "vosk-model-br-0.7",
    "vosk7": "vosk-model-br-0.7",
    "vosk-br-0.8": "vosk-model-br-0.8",
    "vosk8": "vosk-model-br-0.8",
    "vosk-br-0.9": "vosk-model-br-0.9",
    "vosk9": "vosk-model-br-0.9",
    "vosk": get_latest_model(type="vosk"),
}



def is_model_loaded(model_name=None) -> bool:
    model_name = model_name or get_latest_model()
    if model_name in _MODEL_ALIASES:
        model_name = _MODEL_ALIASES[model_name]
    return _loaded_model_name == model_name



def load_model(model_name: str = None) -> Model:
    global _loaded_model_name
    global _loaded_model
    
    if model_name == None:
        if _loaded_model:
            return _loaded_model
        model_name = get_latest_model()
    elif model_name in _MODEL_ALIASES:
        model_name = _MODEL_ALIASES[model_name]
    
    if model_name == _loaded_model_name:
        return _loaded_model

    if _is_valid_vosk_model(model_name):
        # Given path to a model on local storage
        model_path = model_name
    elif model_name in get_local_models():
        # Model is already downloaded
        model_path = os.path.join(_get_model_directory(), model_name)
    elif model_name in get_all_models():
        # Model needs to be downloaded
        model_path = _download(model_name, _get_model_directory())
    else:
        raise RuntimeError(
            f"Model {model_name} is not a valid model; available models = {get_all_models()}"
        )

    print(f"Loading {os.path.basename(model_path.rstrip(os.path.sep))}", file=sys.stderr)
    SetLogLevel(-1)
    _loaded_model = Model(model_path)
    _loaded_model_name = model_name

    return _loaded_model



def _download(model_name: str, root: str) -> str:
    """ Code modified from https://github.com/openai/whisper
        Get the requested model path on disk or download it if not present
    """
    os.makedirs(root, exist_ok=True)
    
    model_path = os.path.join(root, model_name)

    for model in _model_list:
        if model["name"] == model_name:
            url = model["url"]
            break
    else:
        raise RuntimeError("Couldn't find requested model url")
    
    download_target = os.path.join(root, os.path.basename(url))

    print(f"Downloading model from {url}", file=sys.stderr)
    with urllib.request.urlopen(url, context=_certifi_context) as source, open(download_target, "wb") as output:
        with tqdm(
            total=int(source.info().get("Content-Length")),
            ncols=80,
            unit="iB",
            unit_scale=True,
            unit_divisor=1024,
        ) as loop:
            while True:
                buffer = source.read(8192)
                if not buffer:
                    break

                output.write(buffer)
                loop.update(len(buffer))
    
    with zipfile.ZipFile(download_target, 'r') as zip_ref:
        zip_ref.extractall(root)

    os.remove(download_target)

    return model_path



def _is_valid_vosk_model(dir) -> bool:
    if not os.path.isdir(dir):
        return False
    
    vosk_model_files = {
        "final.dubm",
        "final.ie",
        "final.mat",
        "final.mdl",
        "global_cmvn.stats",
        "mfcc.conf",
        "online_cmvn.conf",
        "splice.conf",
        "word_boundary.int"
    }

    files = set()
    for item in os.walk(dir):
        files.update(item[2])
    return vosk_model_files.issubset(files)
